/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


#include "SignedDistFunctors.h"
#include "PhaseFieldConvert.h"

#include "HL_SignedDistTraverse.h"
#include "Recovery.h"
#include "Expressions.h"

namespace AMDiS { namespace base_problems {

namespace detail {
    
  template<typename P>
  CahnHilliard<P>::CahnHilliard(const std::string &name_) 
    : super(name_),
      useMobility(false),
      useReinit(false),
      doubleWell(0),
      gamma(1.0),
      eps(0.1),
      minusEps(-0.1),
      epsInv(10.0),
      minusEpsInv(-10.0),
      epsSqr(0.01),
      minusEpsSqr(-0.01)
  {
    // parameters for CH
    Parameters::get(name_ + "->use mobility", useMobility); // mobility
    Parameters::get(name_ + "->gamma", gamma); // mobility
    Parameters::get(name_ + "->epsilon", eps); // interface width
    
    // type of double well: 0= [0,1], 1= [-1,1]
    Parameters::get(name_ + "->double-well type", doubleWell); 
    
    Parameters::get(name_ + "->use reinit", useReinit);

    // transformation of the parameters
    minusEps = -eps;
    epsInv = 1.0/eps;
    minusEpsInv = -epsInv;
    epsSqr = sqr(eps);
    minusEpsSqr = -epsSqr;
  }


  template<typename P>
  void CahnHilliard<P>::solveInitialProblem(AdaptInfo *adaptInfo) 
  {
    using namespace AMDiS::extensions;
    Flag initFlag = self::initDataFromFile(adaptInfo);
    
    DOFVector<double>* c = self::prob->getSolution()->getDOFVector(1);

    if (!initFlag.isSet(DATA_ADOPTED)) {
      int initialInterface = 0;
      Parameters::get(self::name + "->initial interface", initialInterface);
      double initialEps = eps;
      Parameters::get(self::name + "->initial epsilon", initialEps);

      if (initialInterface == 0) {
	/// horizontale Linie
	double a= 0.0, dir= -1.0;
	Parameters::get(self::name + "->line->pos", a);
	Parameters::get(self::name + "->line->direction", dir);
	c->interpol(new Plane(a, dir));
      }
      else if (initialInterface == 1) {
	/// schraege Linie
	double theta = m_pi/4.0;
	c->interpol(new PlaneRotation(0.0, theta, 1.0));
	transformDOFInterpolation(c, new PlaneRotation(0.0, -theta, -1.0), new AMDiS::Min<double>);
      }
      else if (initialInterface == 2) {
	/// Ellipse
	double a= 1.0, b= 1.0;
	Parameters::get(self::name + "->ellipse->a", a);
	Parameters::get(self::name + "->ellipse->b", b);
	c->interpol(new Ellipse(a,b));
      }
      else if (initialInterface == 3) {
	/// zwei horizontale Linien
	double a= 0.75, b= 0.375;
	Parameters::get(self::name + "->lines->pos1", a);
	Parameters::get(self::name + "->lines->pos2", b);
	c->interpol(new Plane(a, -1.0));
	transformDOFInterpolation(c, new Plane(b, 1.0), new AMDiS::Max<double>);
      }
      else if (initialInterface == 4) {
	/// Kreis
	double radius= 1.0;
	Parameters::get(self::name + "->kreis->radius", radius);
	c->interpol(new Circle(radius));
      } else if (initialInterface == 5) {
	/// Rechteck
	double width = 0.5;
	double height = 0.3;
	WorldVector<double> center; center.set(0.5);
	Parameters::get(self::name + "->rectangle->width", width);
	Parameters::get(self::name + "->rectangle->height", height);
	Parameters::get(self::name + "->rectangle->center", center);
	c->interpol(new Rectangle(width, height, center));
      }
      
      if (useReinit) {
	FiniteElemSpace* feSpace = FiniteElemSpace::provideFeSpace(
	  const_cast<DOFAdmin*>(self::getMesh()->getVertexAdmin()),
	  Lagrange::getLagrange(self::getMesh()->getDim(), 1),
	  self::getMesh(),
	  "P1");
	DOFVector<double> tmp(feSpace, "tmp");
	tmp.interpol(c);
      
	HL_SignedDistTraverse reinit("reinit", self::getMesh()->getDim());
	reinit.calcSignedDistFct(adaptInfo, &tmp);
      
  #ifndef HAVE_PARALLEL_DOMAIN_AMDIS
	Recovery recovery(L2_NORM, 1);
	recovery.recoveryUh(&tmp, *c);
  #else
	c->interpol(&tmp);
  #endif
      }
      

      /// create phase-field from signed-dist-function
      if (doubleWell == 0) {
	forEachDOF(c, new SignedDistToPhaseField(initialEps));
      } else {
	forEachDOF(c, new SignedDistToCh(initialEps));
      }
    }
  }

  template<typename P>
  void CahnHilliard<P>::fillOperators()
  {  
    const FiniteElemSpace* feSpace0 = self::prob->getFeSpace(0);
    const FiniteElemSpace* feSpace1 = self::prob->getFeSpace(1);
    
    DOFVector<double>* c = self::prob->getSolution()->getDOFVector(1);

    // -nabla*(grad(c))
    Operator *opChL = new Operator(feSpace0, feSpace1);
    addSOT(opChL, 1.0);
    
    // div(M(c)grad(mu)), with M(c)=gamma/4*(c^2-1)^2
    Operator *opChLM = new Operator(feSpace1, feSpace0);
    if (useMobility) {
      if (doubleWell == 0)
	addSOT(opChLM, function_(MobilityCH0_(gamma), valueOf(c)) );
      else
	addSOT(opChLM, function_(MobilityCH1_(gamma), valueOf(c)) );
    } else
      addSOT(opChLM, gamma );
    
    // -2*c_old^3 + 3/2*c_old^2
    Operator *opChWExpl = new Operator(feSpace0, feSpace1);
    Operator *opChWImpl = new Operator(feSpace0, feSpace1);
    if (doubleWell == 0) {
      addZOT(opChWExpl, -2.0*pow<3>(valueOf(c)) + 1.5*pow<2>(valueOf(c)));
      addZOT(opChWImpl, -3.0*pow<2>(valueOf(c)) + 3.0*valueOf(c) - 0.5);
    } else {
      addZOT(opChWExpl, -2.0*pow<3>(valueOf(c)));
      addZOT(opChWImpl, -3.0*pow<2>(valueOf(c)) + 1.0);
    }

    
    // mu + eps^2*laplace(c) + c - 3*(c_old^2)*c = -2*c_old^3 [+ BC]
    // ----------------------------------------------------------------------
    self::prob->addMatrixOperator(*opChWImpl, 0, 1);          			/// < -3*phi*c*c_old^2 , psi >
    self::prob->addMatrixOperator(*opChL, 0, 1, &minusEpsSqr, &minusEpsSqr);   	/// < -eps^2*phi*grad(c) , grad(psi) >
    
    // . . . vectorOperators . . . . . . . . . . . . . . .
    self::prob->addVectorOperator(*opChWExpl, 0);            			/// < -2*phi*c_old^3 , psi >

    Operator *opM01new = new Operator(feSpace0, feSpace0);
    addZOT(opM01new, 1.0);
    self::prob->addMatrixOperator(*opM01new, 0, 0);              			/// < phi*mu , psi >
    
    // dt(c) = laplace(mu) - u*grad(c)
    // -----------------------------------
    Operator *opM10new = new Operator(feSpace1, feSpace1);
    addZOT(opM10new, 1.0);
    self::prob->addMatrixOperator(*opM10new, 1, 1); 				/// < phi*c/tau , psi >
    self::prob->addMatrixOperator(*opChLM, 1, 0, self::getTau(), self::getTau());	/// < phi*grad(mu) , grad(psi) >
    
    // . . . vectorOperators . . . . . . . . . . . . . . .
    Operator *opChMold = new Operator(feSpace1, feSpace1);
    addZOT(opChMold, valueOf(c));
    self::prob->addVectorOperator(*opChMold, 1);   				/// < phi*c^old/tau , psi >
  }


  template<typename P>
  void CahnHilliard<P>::finalizeData()
  {
  //   self::setAssembleMatrixOnlyOnce_butTimestepChange(0,1);
  //   self::setAssembleMatrixOnlyOnce_butTimestepChange(1,0);
  //   if (!useMobility)
  //     self::setAssembleMatrixOnlyOnce_butTimestepChange(1,1);
  }

} // end namespace detail

} }
