/** \file CouplingBaseProblem.h */

#ifndef COUPLING_BASE_PROBLEM_H
#define COUPLING_BASE_PROBLEM_H

#include "AMDiS.h"

// coupling structures
#include "CouplingIterationInterface.h"
#include "CouplingTimeInterface.h"
#include "CouplingProblemStat.h"
#include "BaseProblem.h"
#include "ProblemInstat.h"

namespace AMDiS { namespace base_problems {

/**
  * \ingroup Problem 
  *
  * \brief
  */
template<class ProblemType=ProblemStat, class BaseProblemType=BaseProblem<ProblemStat> >
class CouplingBaseProblem : public CouplingIterationInterface,
			   public CouplingTimeInterface,
			   public detail::CouplingProblemStat<ProblemType>
{
public:
  typedef detail::CouplingProblemStat<ProblemType> CProblemStat;

  CouplingBaseProblem(std::string name_, 
		     BaseProblemType *prob0_,
		     BaseProblemType *prob1_=NULL, 
		     BaseProblemType *prob2_=NULL, 
		     BaseProblemType *prob3_=NULL, 
		     BaseProblemType *prob4_=NULL)
  : CProblemStat(name_), name(name_)
  {
    dow = Global::getGeo(WORLD);
    baseProblems.push_back(prob0_);
    if (prob1_) baseProblems.push_back(prob1_);
    if (prob2_) baseProblems.push_back(prob2_);
    if (prob3_) baseProblems.push_back(prob3_);
    if (prob4_) baseProblems.push_back(prob4_);
  }

  ~CouplingBaseProblem() { }  
  
  void initialize(Flag initFlag,
		  ProblemStatSeq *adoptProblem = NULL,
		  Flag adoptFlag = INIT_NOTHING) override
  {  
    for (size_t i = 0; i < baseProblems.size(); i++) {
      for (size_t j = 0; j < baseProblems[i]->getNumProblems(); j++)
	this->addProblem(dynamic_cast<ProblemType*>(baseProblems[i]->getProblem(j)));
      
      addIterationInterface(baseProblems[i]);
      addTimeInterface(baseProblems[i]);
    }

    // initialize all ProblemStat
    CProblemStat::initialize(initFlag, adoptProblem, adoptFlag);
    
    // set mesh dimension
    dim = CProblemStat::getMesh(0)->getDim();
  }  
  
  virtual void initData() {}  
  virtual void finalizeData() {}  
  
  virtual void initTimeInterface()
  {
    for (size_t i = 0; i < baseProblems.size(); i++)
      baseProblems[i]->initData();
    initData();
    
    for (size_t i = 0; i < baseProblems.size(); i++)
      baseProblems[i]->fillOperators();
    fillCouplingOperators();
    
    for (size_t i = 0; i < baseProblems.size(); i++)
      baseProblems[i]->fillBoundaryConditions();
    fillCouplingBoundaryConditions();
    
    for (size_t i = 0; i < baseProblems.size(); i++)
      baseProblems[i]->initData();
    finalizeData();
  }
  
  
  virtual void fillCouplingOperators() {}
  virtual void fillCouplingBoundaryConditions() {}
  
  
  /// get the j-th solution-vector of the i-th problem
  DOFVector<double> *getSolution(int i, int j)
  { FUNCNAME("CouplingBaseProblem::getSolution(i,j)");
    TEST_EXIT(0<=i && 0<=j && i<baseProblems.size() && j<=baseProblems[i]->getNumComponents())
      ("Indices out of range!\n");
    return baseProblems[i]->getSolution()->getDOFVector(j);
  }
  
  
  /// pointer to the j-th feSpace of the i-th problem
  inline const FiniteElemSpace* getFeSpace(int i, int j=0) 
  { FUNCNAME("CouplingBaseProblem::getFeSpace(i,j)");
    TEST_EXIT(0<=i && 0<=j && i<baseProblems.size() && j<=baseProblems[i]->getNumComponents())
      ("Indices out of range!\n");
    return baseProblems[i]->getFeSpace(j); 
  }
  
  std::string getName() { return name; }

protected:
  std::vector<BaseProblemType*> baseProblems;

  unsigned dim;	// dimension of the meshes
  unsigned dow;	// dimension of the world
  
  std::string name;
};

} }

#endif // COUPLING_BASE_PROBLEM_H
