/** \file CoupledBaseProblem_RB.h */

#ifndef COUPLED_BASE_PROBLEM_RB_H
#define COUPLED_BASE_PROBLEM_RB_H

#include "AMDiS.h"

// coupling structures
#include "CouplingIterationInterface.h"
#include "CouplingTimeInterface.h"
#include "CouplingProblemStat.h"
#include "BaseProblem_RB.h"
#include "CouplingBaseProblem.h"
#include "time/ExtendedRosenbrockStationary.h"

namespace AMDiS { namespace base_problems {

/**
  * \ingroup Problem 
  *
  * \brief
  */
class CouplingBaseProblem_RB : public CouplingBaseProblem<ExtendedRosenbrockStationary, BaseProblem_RB>
{
public:
  typedef CouplingBaseProblem<ExtendedRosenbrockStationary, BaseProblem_RB> super;
  typedef BaseProblem_RB BaseProblemType;
  typedef CouplingProblemStatImpl<ExtendedRosenbrockStationary> CProblemStat;

  CouplingBaseProblem_RB(std::string name_, 
			BaseProblemType *prob0_,
			BaseProblemType *prob1_=NULL, 
			BaseProblemType *prob2_=NULL, 
			BaseProblemType *prob3_=NULL, 
			BaseProblemType *prob4_=NULL)
  : super(name_,prob0_,prob1_,prob2_,prob3_,prob4_)
  {
    int componentShift = 0;
    for (size_t i = 0; i < baseProblems.size(); i++) {
      baseProblems[i]->setProblem(new ExtendedRosenbrockStationary(baseProblems[i]->getName() + "->space", componentShift));
      componentShift += baseProblems[i]->getNumComponents();
    }
  }

  ~CouplingBaseProblem_RB() { }  
  
  /// get the j-th stage-solution-vector of the i-th problem
  DOFVector<double> *getStageSolution(int i, int j)
  { FUNCNAME("CouplingBaseProblem_RB::getStageSolution(i,j)");
  
    TEST_EXIT(0<=i && 0<=j && i<baseProblems.size() && j<=baseProblems[i]->getNumComponents())
      ("Indices out of range!\n");
    return baseProblems[i]->getStageSolution(j);
  }  
  
  /// get the j-th unVec-vector of the i-th problem
  DOFVector<double> *getUnVec(int i, int j=0) 
  { FUNCNAME("CouplingBaseProblem_RB::getUnVec(i,j)");
  
    TEST_EXIT(0<=i && 0<=j && i<baseProblems.size() && j<=baseProblems[i]->getNumComponents())
      ("Indices out of range!\n");
    return baseProblems[i]->getUnVec(j);
  }
  
  virtual double estimateTimeError(AdaptInfo* adaptInfo)
  {
    double errorEst = 0.0;
    for (size_t i = 0; i < CProblemStat::getNumProblems(); i++)
      errorEst = std::max(errorEst, CProblemStat::getProblem(i)->estimateTimeError(adaptInfo));
    
    adaptInfo->setTimeEst(errorEst);
    return errorEst;
  }
  
  double getNewTimestep(double tol, double tau, bool restrict = true)
  {
    double newTimestep = 1.e15;
    for (size_t i = 0; i < baseProblems.size(); i++)
      newTimestep = std::min(newTimestep, baseProblems[i]->getNewTimestep(tol, tau, restrict));
    return newTimestep;
  }
  
  double getNewTimestep(double tol, double oldTau, double tau, bool restrict = true)
  {
    double newTimestep = 1.e15;
    for (size_t i = 0; i < baseProblems.size(); i++)
      newTimestep = std::min(newTimestep, baseProblems[i]->getNewTimestep(tol, oldTau, tau, restrict));
    return newTimestep;
  }

  // Rosenbrock methods
  //_________________________________________________________________________

  void acceptTimestep(AdaptInfo* adaptInfo)
  {
    for (size_t i = 0; i < baseProblems.size(); i++)
      baseProblems[i]->acceptTimestep(adaptInfo);
  }

//   void reset()
//   {
//     for (size_t i = 0; i < CProblemStat::getNumProblems(); i++)
//       CProblemStat::getProblem(i)->reset();
//     for (size_t i = 0; i < baseProblems.size(); i++)
//       baseProblems[i]->reset();
//   }

  void setTau(double *ptr)
  {
    for (size_t i = 0; i < CProblemStat::getNumProblems(); i++)
      CProblemStat::getProblem(i)->setTau(ptr);
  }
  
  void setOldTime(double t)
  {
    for (size_t i = 0; i < CProblemStat::getNumProblems(); i++)
      CProblemStat::getProblem(i)->setOldTime(t);
  }
  
  void setTime(AdaptInfo* adaptInfo)
  {
    for (size_t i = 0; i < baseProblems.size(); i++)
      baseProblems[i]->setTime(adaptInfo);
  }
	
  void reduceOrder(double tauRej, double tau)
  {
    for (size_t i = 0; i < baseProblems.size(); i++)
      baseProblems[i]->reduceOrder(tauRej, tau);
  }
  
  double* getOldTimePtr()
  {
    return CProblemStat::getProblem(0)->getOldTimePtr();
  }
};

} }

#endif // COUPLED_BASE_PROBLEM_RB_H
