/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
#ifndef NAVIER_STOKES_TAYLOR_HOOD_MULTIPHASE_H
#define NAVIER_STOKES_TAYLOR_HOOD_MULTIPHASE_H

#include "NavierStokes_TaylorHood.h"

namespace AMDiS { namespace base_problems {

namespace detail {
  
  /** \ingroup NavierStokes_TaylorHood
  * \brief
  * Navier-Stokes multi-phase problem, using Taylor Hood elements
  * Standard implementation for two phases, but can be extended to
  * many more, by defining the multiPhase variable and overloading
  * the initTimestep, where densityPhase and viscosityPhase are defined
  * depending on the phases
  */
  template<typename ProblemType>
  class NavierStokes_TH_MultiPhase : public NavierStokes_TaylorHood<ProblemType>
  {
  public: // typedefs

    typedef NavierStokes_TaylorHood<ProblemType> super;
    typedef NavierStokes_TH_MultiPhase<ProblemType> self;

  public: // methods

    NavierStokes_TH_MultiPhase(const std::string &name_, bool createProblem = true);

    /// deletes multiPhase, viscosityPhase and densityPhase
    ~NavierStokes_TH_MultiPhase();

    /** definition of constant multiPhase and const
    * viscosity/density DOFVectors,
    * initTimeInterface of super
    **/
    virtual void initData() override;
    virtual void transferInitialSolution(AdaptInfo *adaptInfo) override;

    /** initTimestep of super and
    * initialization of densityPhase and viscosityPhase
    **/
    virtual void initTimestep(AdaptInfo *adaptInfo) override;
    virtual void closeTimestep(AdaptInfo *adaptInfo) override;

    /** pointer to the multiPhase DOFVector, that
    * indicates the different phases. Will be initialized
    * in \ref initTimeInterface with const 1
    **/
    void setMultiPhase(DOFVector<double>* p) { multiPhase=p; }

    virtual void fillOperators() override;

    virtual void addLaplaceTerm(int i) override;
    
    // get-methods
    
    DOFVector<double>* getViscosityPhase()
    {
      return viscosityPhase;
    }
    
    DOFVector<double>* getDensityPhase()
    {
      return densityPhase;
    }
    
    double getViscosity(int i=0)
    {
      if (i == 0)
	return viscosity1;
      else
	return viscosity2;
    }
    
    double getDensity(int i=0)
    {
      if (i == 0)
	return density1;
      else
	return density2;
    }
    
    void rejectTimestep()
    {
      *stepSolution = *oldSolution;
    }
    
    void acceptTimestep()
    { FUNCNAME("NavierStokes_TH_MultiPhase::acceptTimestep");
      MSG("Timestep accepted.\n");
      *oldSolution = *self::prob->getSolution();
      *stepSolution = *self::prob->getSolution();
    }
    
    SystemVector* getStepSolution()
    {
      return stepSolution;
    }

  protected: // variables

    ///viscosity of phase 1
    double viscosity1;
    
    ///viscosity of phase 2
    double viscosity2;
    
    ///density of phase 1
    double density1;
    
    ///density of phase 2
    double density2;
    
    int phaseFieldType;

    /// phase dependent density
    DOFVector<double> *densityPhase;
    
    /// phase dependent viscosity
    DOFVector<double> *viscosityPhase;
    
    /// phase inticator
    DOFVector<double> *multiPhase;
    
    SystemVector* oldSolution;
    SystemVector* stepSolution;

  };


  /** linear interpolation between two values (like density, viscosity)
  *  using a phase-field variable in [-1,1]
  **/
  struct LinearInterpolation1 : public FunctorBase
  {  
    typedef double result_type;
    int getDegree(int d0) const { return 1; }

    LinearInterpolation1(double c1, double c2) 
    {
      a = (c1-c2)/2.0; 
      b = (c1+c2)/2.0; 
      cmin = std::min(c1,c2); 
      cmax = std::max(c1,c2);
    }
      
    double operator()(const double &phase) const 
    { 
      double result = b+a*phase;
      if (result < cmin) result = cmin; 
      if (result > cmax) result = cmax;
      return result;
    }

  private:
    double a,b,cmin,cmax;
  };
}

typedef base_problems::detail::NavierStokes_TH_MultiPhase<AMDiS::ProblemStat> NavierStokes_TH_MultiPhase;

} }

#include "NavierStokes_TH_MultiPhase.hh"

#endif // NAVIER_STOKES_TAYLOR_HOOD_MULTIPHASE_H
