/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
#include "NavierStokes_TaylorHood_RB.h"

namespace AMDiS { namespace base_problems {

NavierStokes_TaylorHood_RB::NavierStokes_TaylorHood_RB(const std::string &name_, bool createProblem) :
  super(name_, createProblem),
  velocity(NULL),
  laplaceType(0),
  oldTimestep(0.0),
  viscosity(1.0),
  fileWriter(NULL)
{  
  // fluid viscosity
  Initfile::get(name + "->viscosity", viscosity);

  // type of laplace operator: 
  //   0... div(nu*grad(u)), 
  //   1... div(0.5*nu*(grad(u)+grad(u)^T))
  Initfile::get(name + "->laplace operator", laplaceType); 

  // volume force
  force.set(0.0);
  Initfile::get(name + "->force", force);
}


NavierStokes_TaylorHood_RB::~NavierStokes_TaylorHood_RB()
{
  if (velocity != NULL) {
    delete velocity; velocity = NULL;
  }
  if (fileWriter) {
    delete fileWriter; fileWriter = NULL;
  }
}


void NavierStokes_TaylorHood_RB::initData()
{
  if (velocity == NULL)
    velocity = new DOFVector<WorldVector<double> >(getFeSpace(0), "velocity");
  fileWriter = new FileVectorWriter(name + "->velocity->output", getFeSpace()->getMesh(), velocity);

  super::initData();
}


void NavierStokes_TaylorHood_RB::transferInitialSolution(AdaptInfo *adaptInfo)
{
  calcVelocity();

  fileWriter->writeFiles(adaptInfo, true);
  writeFiles(adaptInfo, true);

  // initial parameters for detecting mesh changes
  oldMeshChangeIdx= getMesh()->getChangeIndex();
}


void NavierStokes_TaylorHood_RB::fillOperators()
{
  using namespace extensions;
  
  WorldVector<DOFVector<double>*> U;
  WorldVector<DOFVector<double>*> U_old;
  for (unsigned i = 0; i < dow; ++i) {
    U[i] = prob->getStageSolution(i);
    U_old[i] = prob->getUnVec(i);
  }
  DOFVector<double> *p = prob->getStageSolution(dow);
// 		    *p_old = prob->getUnVec(dow);
  
  // fill operators for prob
  for (unsigned i = 0; i < dow; ++i) {
    /// < d_t(u_i) , psi >
    prob->addTimeOperator(i, i);
 
    /// < u*grad(u_i) , psi >
    Operator *opUGradU = new Operator(getFeSpace(i), getFeSpace(i));
    opUGradU->addTerm(new WorldVec_FOT(U, -1.0), GRD_PHI);
    opUGradU->setUhOld(U[i]);
    prob->addVectorOperator(*opUGradU, i);

    Operator *opUGradV = new Operator(getFeSpace(i), getFeSpace(i));
    opUGradV->addTerm(new WorldVec_FOT(U_old, -1.0), GRD_PHI);
    prob->addMatrixOperator(*opUGradV, i, i, &minus1, &minus1);

    for (unsigned j = 0; j < dow; ++j) {
      Operator *opVGradU = new Operator(getFeSpace(i), getFeSpace(j));
      opVGradU->addTerm(new PartialDerivative_ZOT(U_old[i], j, -1.0));
      prob->addMatrixOperator(*opVGradU, i, j, &minus1, &minus1);
    }

    addLaplaceTerm(i);
  
    /// < p , d_i(psi) >
    Operator *opGradP = new Operator(getFeSpace(i),getFeSpace(dow));
    opGradP->addTerm(new PartialDerivative_FOT(i, -1.0), GRD_PSI);
    opGradP->setUhOld(p);
    prob->addVectorOperator(*opGradP, i);
    prob->addMatrixOperator(*opGradP, i, dow, &minus1, &minus1);
  
    /// external force, i.e. gravitational force
    if (std::abs(force[i]) > DBL_TOL) {
      Operator *opForce = new Operator(getFeSpace(i), getFeSpace(i));
      opForce->addTerm(new Simple_ZOT(force[i]));
      prob->addVectorOperator(*opForce, i);
    }
  }

  /// div(u) = 0
  for (unsigned i = 0; i < dow; ++i) {
    /// < d_i(u'_i) , psi >
    Operator *opDivU = new Operator(getFeSpace(dow),getFeSpace(i));
    opDivU->addTerm(new PartialDerivative_FOT(i), GRD_PHI);
    opDivU->setUhOld(U[i]);
    prob->addVectorOperator(*opDivU, dow);
    prob->addMatrixOperator(*opDivU, dow, i, &minus1, &minus1);
  }
}


void NavierStokes_TaylorHood_RB::addLaplaceTerm(int i)
{
  using namespace extensions;
  
  /// < alpha*[grad(u)+grad(u)^t] , grad(psi) >
  if (laplaceType == 1) {
    for (unsigned j = 0; j < dow; ++j) {
      Operator *opLaplaceUi1 = new Operator(getFeSpace(i), getFeSpace(j));
      opLaplaceUi1->addTerm(new MatrixIJ_SOT(j, i, viscosity));
      opLaplaceUi1->setUhOld(prob->getStageSolution(j));
      prob->addVectorOperator(*opLaplaceUi1, i, &minus1, &minus1);
      prob->addMatrixOperator(*opLaplaceUi1, i, j);
    }
  }
    
  /// < alpha*grad(u'_i) , grad(psi) >
  Operator *opLaplaceUi = new Operator(getFeSpace(i), getFeSpace(i));
    opLaplaceUi->addTerm(new Simple_SOT(viscosity));
  opLaplaceUi->setUhOld(prob->getStageSolution(i));
  prob->addVectorOperator(*opLaplaceUi, i, &minus1, &minus1);
  prob->addMatrixOperator(*opLaplaceUi, i, i);
}


void NavierStokes_TaylorHood_RB::closeTimestep(AdaptInfo *adaptInfo)
{
  calcVelocity();
  fileWriter->writeFiles(adaptInfo, false);
  writeFiles(adaptInfo, false);
}

} }
