#ifndef CAHN_HILLIARD_PROBLEM_STAT_H
#define CAHN_HILLIARD_PROBLEM_STAT_H

#include "ProblemStat.h"
#include "Expressions.h"
#include "time/RosenbrockStationary.h"

using namespace AMDiS;

struct CahnHilliardProblemStat : ProblemStat
{
  CahnHilliardProblemStat(std::string name)
    : ProblemStat(name)
  { }
  
  template< typename Data >
  void setData(const Data* data)
  {
    // not data needed
  }
  
  void fillOperators()
  {
    DOFVector<double>* c = getSolution(1);
    const FiniteElemSpace* feSpace = c->getFeSpace();
  
    /// M := (phi, psi)
    Operator *mass_op = new Operator(feSpace);	     
    addZOT(mass_op, 1.0);
    addMatrixOperator(*mass_op, 0,0);
    setAssembleMatrixOnlyOnce(0,0,true);
    
    /// L := (grad(phi), grad(psi))
    Operator *laplace_op = new Operator(feSpace);
    addSOT(laplace_op, 1.0);
    addMatrixOperator(*laplace_op, 1,1);
    setAssembleMatrixOnlyOnce(1,1,true);
    
    /// N := (-2*c^2 * phi, psi)
    Operator *rho3_op2 = new Operator(feSpace);
    addZOT(rho3_op2, -3.0 * pow<2>(valueOf(c)));
    addMatrixOperator(*rho3_op2, 1,0);
    setAssembleMatrixOnlyOnce(1,0,false);
    
    /// K := (M*grad(phi), grad(psi))
    Operator *mobility_op_ch = new Operator(feSpace);
    addSOT(mobility_op_ch, 1.0);
    addMatrixOperator(*mobility_op_ch, 0,1);
    setAssembleMatrixOnlyOnce(0,1,true);
    
    // --- vector operators ----------------------------------------------------
    
    /// (-c^3, psi)
    Operator *rho3_op_rhs = new Operator(feSpace);
    addZOT(rho3_op_rhs, -2.0 * pow<3>(valueOf(c)));
    addVectorOperator(*rho3_op_rhs, 0);
    
    /// (c, psi)
    Operator *mass_op_rhs = new Operator(feSpace);
    addZOT(mass_op_rhs, valueOf(c));
    addVectorOperator(*mass_op_rhs, 1);
  }
};

  
// =============================================================================
  
/// TODO: create the Rosenbrock operators
struct CahnHilliardProblemStatRB : RosenbrockStationary
{
  CahnHilliardProblemStatRB(std::string name)
    : RosenbrockStationary(name)
  { }
  
  template< typename Data >
  void setData(const Data* data)
  {
    // no data needed
  }
  
  void buildAfterCoarsen(AdaptInfo *adaptInfo, Flag flag,
	  	         bool asmMatrix, bool asmVector)
  {
    ProblemStat::buildAfterCoarsen(adaptInfo, flag, asmMatrix, asmVector);
  }

  void fillOperators()
  {
    DOFVector<double>* mu_old = getUnVec(0);
    DOFVector<double>* c_old = getUnVec(1);
    
    DOFVector<double>* mu_s = getStageSolution(0);
    DOFVector<double>* c_s = getStageSolution(1);
    
    const FiniteElemSpace* feSpace = c_old->getFeSpace();
  
    /// (phi, psi)
    Operator *mass_op = new Operator(feSpace);	     
    addZOT(mass_op, 1.0);
    addMatrixOperator(*mass_op, 0,0);
    setAssembleMatrixOnlyOnce(0,0,true);
    
    /// (grad(phi), grad(psi))
    Operator *laplace_op = new Operator(feSpace);
    addSOT(laplace_op, 1.0);
    addMatrixOperator(*laplace_op, 1,1);
    setAssembleMatrixOnlyOnce(1,1,true);
    
  }
  
  SystemVector& getUnVec()
  {
    return *unVec;
  }
  
  DOFVector<double>* getUnVec(int i)
  {
    return unVec->getDOFVector(i);
  }
  
  SystemVector& getNewUn()
  {
    return *newUn;
  }
  
  DOFVector<double>* getNewUn(int i)
  {
    return newUn->getDOFVector(i);
  }
  
  SystemVector& getTmpVec()
  {
    return *tmp;
  }
  
  SystemVector& getLowSol()
  {
    return *lowSol;
  }
  
  DOFVector<double>* getLowSol(int i)
  {
    return lowSol->getDOFVector(i);
  }

  SystemVector& getStageSolution()
  {
    return *stageSolution;
  }

  DOFVector<double>* getStageSolution(int i)
  {
    return stageSolution->getDOFVector(i);
  }

  SystemVector& getStageSolutions(int i)
  {
    return *(stageSolutions[i]);
  }

  SystemVector& getTimeRhsVec()
  {
    return *timeRhsVec;
  }

  DOFVector<double>* getTimeRhsVec(int i)
  {
    return timeRhsVec->getDOFVector(i);
  }
  
private:

};

#endif // CAHN_HILLIARD_PROBLEM_STAT_H