#include <limits>
#include "Refinement.h"

using namespace std;
using namespace AMDiS;


Refinement::Refinement(DOFVector<double> *phasefield_, 
		       AbstractFunction<double, WorldVector<double> > *phasefieldFct_, 
		       RefinementManager *refinementManager_, 
		       CoarseningManager *coarseningManager_) 
  : phasefield(phasefield_), 
    phasefieldFct(phasefieldFct_), 
    refinementManager(refinementManager_), 
    coarseningManager(coarseningManager_),
    info_(0)
{
  feSpace = phasefield->getFeSpace();
  mesh = feSpace->getMesh();
}


Refinement::~Refinement() 
{
}


void Refinement::refine(double h)
{
  FUNCNAME("Refinement::refine()");
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
  MPI::COMM_WORLD.Barrier();
  double first = MPI::Wtime();
  double m0 = 0.0, m1 = 0.0;
#endif
  // build mesh for phasefield-function
  bool meshChanged=true;
  Flag markFlag;
  double minH = 0.0, maxH = 1.0, oldMinH = 100.0, oldMaxH = 0.0;

  int i = 0;
  while (meshChanged || minH != oldMinH || maxH != oldMaxH) {
    oldMinH = minH; 
    oldMaxH = maxH;

    if (phasefieldFct != NULL)
      phasefield->interpol(phasefieldFct);

#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    double t0 = MPI::Wtime();
    markElements(markFlag, h);
    m0 += MPI::Wtime() - t0;

    t0 = MPI::Wtime();
    meshChanged = refineMesh(markFlag);
    m1 += MPI::Wtime() - t0;
    
#else
    markElements(markFlag, h);
    meshChanged = refineMesh(markFlag);
#endif

    calcMeshSizes(minH, maxH);
      
    if (meshChanged || minH != oldMinH || maxH != oldMaxH) {
      int nr = phasefield->getUsedSize();
      MSG("%d Mesh sizes: [%f, %f], DOFs: %d, ELs: %d\n", i++,
		     minH, maxH, nr, mesh->getNumberOfLeaves());
    }
  }
  
  if (phasefieldFct != NULL)
    phasefield->interpol(phasefieldFct);

#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
  Parallel::mpi::globalMax(m0);
  Parallel::mpi::globalMax(m1);

  MPI::COMM_WORLD.Barrier();
  MSG("Local mesh refinement (phase field) needed %.5f seconds [ %.5f %.5f ]\n", MPI::Wtime() - first, m0, m1);
#endif
}


void Refinement::calcMeshSizes(double &minH, double &maxH)
{
  int dim = mesh->getDim();
  FixVec<WorldVector<double>, VERTEX> coords(dim, NO_INIT);

  TraverseStack stack;
  ElInfo *elInfo = stack.traverseFirst(mesh, -1, Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS);
  minH=1e15; maxH=0.0;
  while (elInfo) {
    coords = elInfo->getCoords();
    double h = 0.0;
    for (int i = 0; i < coords.getSize(); i++)
      for(int j = i + 1; j < coords.getSize(); j++)
	if (i != j) 
	  h = std::max(h, norm(coords[i] - coords[j]));
    minH = std::min(h, minH);
    maxH = std::max(h, maxH);
    elInfo = stack.traverseNext(elInfo);
  }
}


void Refinement::markElements(Flag &markFlag, double hInterface)
{
  DOFVector<double> *vec = phasefield;
  const DOFAdmin* admin = feSpace->getAdmin();
  const BasisFunction *basFcts = feSpace->getBasisFcts();

  int dim = mesh->getDim();
  int numBasFcts = basFcts->getNumber();

  std::vector<DegreeOfFreedom> localIndices(numBasFcts);
  FixVec<WorldVector<double>, VERTEX> coords(dim, NO_INIT);

  bool elMarkRefine=false, elMarkCoarsen=false;

  TraverseStack stack;
  ElInfo *elInfo = stack.traverseFirst(mesh, -1, Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS);
  while (elInfo) {
    Element *el = elInfo->getElement();
    basFcts->getLocalIndices(el, admin, localIndices);
    // for every vertex
    double smax = -1e12, smin = 1e12;
    coords = elInfo->getCoords();
    for (int i = 0; i < numBasFcts; i++) {
      // search for coords of this vertex in the coord-list of this dof 
      smin = std::min(smin, (*vec)[localIndices[i]]);
      smax = std::max(smax, (*vec)[localIndices[i]]);
    }
    
    if (smin < 0.1 && smax > -0.1) {
      double h = 0.0;
      for (int i = 0; i < coords.getSize(); i++)
	for (int j = i + 1; j < coords.getSize(); j++)
	  h = std::max(h, norm(coords[i] - coords[j]));     

      if (h > hInterface) {
	elMarkRefine = true;
	elInfo->getElement()->setMark(1);
      }
    } else {
      if (smax - smin > 0.2) {
	elMarkRefine = true;
	elInfo->getElement()->setMark(1);
      }
      if (smax -smin < 0.1) {
	elMarkCoarsen = true;
	elInfo->getElement()->setMark(-1);  
      }
    }
    
    elInfo = stack.traverseNext(elInfo);
  }

  markFlag = 0;
  if (elMarkRefine) 
    markFlag = 1;
  if (elMarkCoarsen)
    markFlag |= 2;  
}


bool Refinement::refineMesh(Flag markFlag)
{
  FUNCNAME("Refinement::refineMesh()");

  int oldSize = phasefield->getUsedSize();
  if (markFlag.isSet(1))
    refinementManager->refineMesh(mesh);
  if (markFlag.isSet(2))
    coarseningManager->coarsenMesh(mesh);
  if (markFlag.isSet(1) || markFlag.isSet(2)) {
    int newSize = phasefield->getUsedSize();
    if (oldSize != newSize) 
      return true;
    INFO(info_,1)(">> InterfaceRefinement refined DOFs: %d\n",newSize); 
  }
  return false;
}

